/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.assertlevel;

import org.builtu.SynchronizedBy;
import org.builtu.Syncronization;

/**
 * <h1>Level-based assert facility.</h1>
 *
 * <h2>Overview</h2>
 * This facility allows you to perform fine-grained assertions in your code. Consider following:
 * <pre>
 * assert someObject == expected; // trivial assert
 * assert someExpensiveAssert();  // expensive assert
 * </pre>
 * Using standard java tools, you have only 2 options: either turn on all assertions, including expesive ones (which
 * may br TOO expensive), or turn them off completely. This encourages developer not to write expensive assertions
 * at all. 
 * By using this facility, you can assign "threat level" to each and every assert you write, so you can build 
 * defensive code with lots of layers that can be turned on or off at will.
 * 
 * <h2>Threat level</h2>
 * See {@link ThreatLevel}
 * 
 * <h2>Developing code with assertlevel</h2>
 * FIRST RULE: don't write asserts that change state!!!
 * 
 * Use something like:
 * <pre>
 * assert AL.std() || [your condition here]; // same as AL.level({@link ThreatLevel#STANDARD}.code) ||...
 * assert AL.input() || [your condition here]; // same as AL.level({@link ThreatLevel#INPUT_PARAMETERS_CHECK}.code) ||... 
 * assert AL.output() || [your condition here]; // same as AL.level({@link ThreatLevel#OUTPUT_PARAMETERS_CHECK.code}) ||...
 * assert AL.invariant() || [your condition here]; // same as AL.level({@link ThreatLevel#INTERNAL_INVARIANT_CHECK}.code)
 * assert AL.paranoid() || [your condition here]; // same as AL.level({@link ThreatLevel#PARANOID}.code) ||...
 * </pre>
 *
 * In development, run all code with assert level {@link ThreatLevel#PARANOID} (which his default) since it is most...
 * well...paranoid and it contains some internal asserts that tries it's best to check your checking code. 
 * 
 * <h2>Production configuration</h2>
 * There is a single system property that governs this facility: "builtu.assertlevel.level". It can be specified
 * using JRE configuration files or in command line: java -jar myapp.jar -Dbuiltu.assertlevel.level=PARANOID
 * 
 * You can assign strings to this property (see {@link ThreatLevel} enumeration) like this:
 * -Dbuiltu.assertlevel.level=INPUT_PARAMETERS_CHECK 
 * 
 * Or, you can assign specific level code:
 * -Dbuiltu.assertlevel.level=20000
 * 
 * DEFAULT LEVEL IS {@link ThreatLevel#PARANOID} 
 * 
 * @author borisman
 *
 */
public class AL {
	public static final String ASSERTLEVEL_SYSTEM_PROPERTY = "builtu.assertlevel.level";
	public static final ThreatLevel DEFAULT_LEVEL = ThreatLevel.PARANOID;
	
	@SynchronizedBy(synchronization=Syncronization.CLASSLOADER)
	public static final int level;
	static {
		level = readSystemLevel();
	}
	
	private static int readSystemLevel() {
		final String expr = System.getProperty(ASSERTLEVEL_SYSTEM_PROPERTY);
		if (expr == null) return DEFAULT_LEVEL.code;
		final ThreatLevel level = ThreatLevel.fromName(expr.trim().toUpperCase());
		if (level != null) return level.code;
		
		return Integer.parseInt(expr);
	}
	
	public static boolean std() {
		return level(ThreatLevel.STANDARD.code);
	}
	
	public static boolean input() {
		return level(ThreatLevel.INPUT_PARAMETERS_CHECK.code);
	}
	
	public static boolean output() {
		return level(ThreatLevel.OUTPUT_PARAMETERS_CHECK.code);
	}
	
	public static boolean invariant() {
		return level(ThreatLevel.INTERNAL_INVARIANT_CHECK.code);
	}
	
	public static boolean paranoid() {
		return level(ThreatLevel.PARANOID.code);
	}

	public static boolean level(final int lev) {
		return lev > level;
	}

}
