/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.collections.adapters;

import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

/** Adapts list of MUTABLE object to list of IMMUTABLE object list. 
 * Typical usages is when you have a list of objects and want to get a list of (immutable) interfaces of those objects
 * without building a new list 
 * @author borisman
 *
 * @param <IMMUTABLE> immutable interface
 * @param <MUTABLE> mutable interface or class, should extend MUTABLE interface
 */
public class ImmutableListAdapter<IMMUTABLE, MUTABLE extends IMMUTABLE> implements List<IMMUTABLE>, Serializable {

	private List<MUTABLE> backend;
	
	public ImmutableListAdapter(final List<MUTABLE> list) {
		this.backend = list;
	}

	@Override public boolean add(final IMMUTABLE e) {
		throw new UnsupportedOperationException();
	}

	@Override public void add(final int index, final IMMUTABLE element) {
		throw new UnsupportedOperationException();
	}

	@Override public boolean addAll(final Collection<? extends IMMUTABLE> c) {
		throw new UnsupportedOperationException();
	}

	@Override public boolean addAll(final int index, final Collection<? extends IMMUTABLE> c) {
		throw new UnsupportedOperationException();
	}

	@Override public void clear() {
		throw new UnsupportedOperationException();
	}

	@Override public boolean contains(final Object o) {
		return backend.contains(o);
	}

	@Override public boolean containsAll(final Collection<?> c) {
		return backend.containsAll(c);
	}

	@Override public IMMUTABLE get(final int index) {
		return backend.get(index);
	}

	@Override public int indexOf(final Object o) {
		return backend.indexOf(o);
	}

	@Override public boolean isEmpty() {
		return backend.isEmpty();
	}

	@Override public Iterator<IMMUTABLE> iterator() {
		return listIterator();
	}

	@Override public int lastIndexOf(final Object o) {
		return backend.lastIndexOf(o);
	}
	
	private static class ListIteratorAdapter <IMMUTABLE, MUTABLE extends IMMUTABLE>  
		implements ListIterator<IMMUTABLE> 
	{

		private final ListIterator<MUTABLE> backend;
		
		public ListIteratorAdapter(final ListIterator<MUTABLE> backend) {
			this.backend = backend;
		}

		@Override public void add(final IMMUTABLE e) {
			throw new UnsupportedOperationException();
		}

		@Override public boolean hasNext() {
			return backend.hasNext();
		}

		@Override public boolean hasPrevious() {
			return backend.hasPrevious();
		}

		@Override public IMMUTABLE next() {
			return backend.next();
		}

		@Override public int nextIndex() {
			return backend.nextIndex();
		}

		@Override public IMMUTABLE previous() {
			return backend.previous();
		}

		@Override public int previousIndex() {
			return backend.previousIndex();
		}

		@Override public void remove() {
			throw new UnsupportedOperationException();
		}

		@Override public void set(final IMMUTABLE e) {
			throw new UnsupportedOperationException();
		}
	}

	@Override public ListIterator<IMMUTABLE> listIterator() {
		return new ListIteratorAdapter<IMMUTABLE, MUTABLE>(backend.listIterator());
	}

	@Override public ListIterator<IMMUTABLE> listIterator(int index) {
		return new ListIteratorAdapter<IMMUTABLE, MUTABLE>(backend.listIterator(index));
	}

	@Override public boolean remove(final Object o) {
		throw new UnsupportedOperationException();
	}

	@Override public IMMUTABLE remove(final int index) {
		throw new UnsupportedOperationException();
	}

	@Override public boolean removeAll(final Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	@Override public boolean retainAll(final Collection<?> c) {
		throw new UnsupportedOperationException();
	}

	@Override public IMMUTABLE set(final int index, final IMMUTABLE element) {
		throw new UnsupportedOperationException();
	}

	@Override public int size() {
		return backend.size();
	}

	@Override public List<IMMUTABLE> subList(final int fromIndex, final int toIndex) {
		return new ImmutableListAdapter<IMMUTABLE, MUTABLE>(backend.subList(fromIndex, toIndex));
	}

	@Override public Object[] toArray() {
		return backend.toArray();
	}

	@Override public <T> T[] toArray(T[] a) {
		return backend.toArray(a);
	}
	
	private static final long serialVersionUID = 1L;
}
