/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.collections.graph;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.builtu.collections.adapters.ImmutableListAdapter;


/**
 * Directed graph mutable implementation.
 * 
 * NOTE! It is possible to make grapth with cycles using this implementation, but that is not recommended!
 * @author borisman
 *
 * @param <E>
 */
public class NodeImpl<E> implements Serializable, Node<E> {
	private final String nodeId;
	private final List<NodeImpl<E>> parents, children; 
	private final List<Node<E>> parentsRO, childrenRO;
	private final Map<String, NodeImpl<E>> childrenMap;
	
	private E data; 
	
	public NodeImpl (final String nodeId) {
		this(nodeId, 1, 16);
	}

	public NodeImpl(final String nodeId, final int hintNumberOfParents, final int hintNumberOfChildren) {
		checkNodeId(nodeId);
		this.nodeId = nodeId;
		parents = new ArrayList<NodeImpl<E>>(hintNumberOfParents);
		children = new ArrayList<NodeImpl<E>>(hintNumberOfChildren);
		parentsRO = new ImmutableListAdapter<Node<E>, NodeImpl<E>>(parents);
		childrenRO =  new ImmutableListAdapter<Node<E>, NodeImpl<E>>(children);
		childrenMap = new HashMap<String, NodeImpl<E>>(hintNumberOfChildren);
	}

	private static void checkNodeId(final String nodeId) {
		if (nodeId == null) throw new NullPointerException("NodeId cannot be null");
		if (nodeId.indexOf('/')>=0) throw new IllegalArgumentException("NodeId cannot contain '/' (got:'"+nodeId+"')");
	}

	@Override public Node<E> get(final int idx) {
		return children.get(idx);
	}

	@Override public Node<E> get(final String nodeId) {
		checkNodeId(nodeId);
		return childrenMap.get(nodeId);
	}

	@Override public List<Node<E>> getChildren() {
		return childrenRO;
	}

	public List<NodeImpl<E>> getChildrenImpl() {
		return children;
	}
	
	@Override public E getValue() {
		return data;
	}

	public void setValue(final E data) {
		this.data = data;
	}

	@Override public String getNodeId() {
		return nodeId;
	}

	@Override public List<Node<E>> getParents() {
		return parentsRO;
	}

	@Override public Node<E> path(final String path) {
		return pathImpl(path);
	}

	@Override public String getFirstPath() {
		final StringBuilder b = new StringBuilder();
		Node<E> current = this;
		while (current.hasParents()) {
			if (b.length()>0) b.insert(0, '/');
			b.insert(0, current.getNodeId());
			current = current.getParents().get(0);
		}
		return b.toString();
	}
	
	@Override public List<String> getAllPaths() {
		final List<String> res = new ArrayList<String>();
		gatherPaths(res, new StringBuilder(50), this);
		return res;
	}
	

	/**
	 * Syntactic sugar for node.getChildren().size()>0
	 */
	@Override public boolean hasChildren () {
		return getChildren().size()>0;
	}
	
	/**
	 * Syntactic sugar for node.getParents().size()>0
	 */
	@Override public boolean hasParents () {
		return getParents().size()>0;
	}
	
	public List<NodeImpl<E>> getParentsImpl() {
		return parents;
	}

	public NodeImpl<E> getParentImpl(final int idx) {
		return parents.get(0);
	}
	
	public NodeImpl<E> pathImpl(final String path) {
		if (path == null) throw new NullPointerException("Path cannot be null");
		final int len = path.length();
		if (len == 0) return null;
		if (path.charAt(0)=='/' || path.charAt(len-1) == '/') {
			throw new IllegalArgumentException("Path cannot start or end with slash (got:'"+path+"')");
		}
		NodeImpl<E> current = this;
		for (final String s : path.split("/")) {
			current = current.getChildImpl(s);
			if (current == null) break;
		}
		return current;
	}
	
	public NodeImpl<E> getChildImpl(final String nodeId) {
		checkNodeId(nodeId);
		return childrenMap.get(nodeId);
	}

	public void addChild(final NodeImpl<E> child) throws IllegalArgumentException {
		if (this == child) throw new IllegalArgumentException("Can't add myself as child of myself");
		final NodeImpl<E> old = childrenMap.put(child.nodeId, child);
		if (old != null) {
			if (old == child) throw new IllegalArgumentException("Can't add same child twice");
			childrenMap.put(child.nodeId, old);
			throw new IllegalArgumentException("Child with nodeId '"+child.nodeId+"' already added");
		}
		children.add(child);
		child.parents.add(this);
	}

	public boolean removeChild(final NodeImpl<E> child) {
		final NodeImpl<E> old = childrenMap.remove(child.nodeId);
		final boolean res = old != null;
		if (res) {
			assert old == child;
			final boolean removedFromList = children.remove(child);
			assert removedFromList;
			final boolean removedFromChild = child.parents.remove(this);
			assert removedFromChild;
		}
		return res;
	}

	public static<T extends Comparable<T>> void sort (final NodeImpl<T> root) {
		final Comparator<NodeImpl<T>> cmp = new Comparator<NodeImpl<T>> () {
			@Override public int compare(NodeImpl<T> o1, NodeImpl<T> o2) {
				return o1.getValue().compareTo(o2.getValue());
			}};
		sortNodes(root, cmp);
	}
	
	public static<T> void sort (final NodeImpl<T> root, final Comparator<? super T> comparator) {
		final Comparator<NodeImpl<T>> cmp = new Comparator<NodeImpl<T>> () {
			@Override public int compare(NodeImpl<T> o1, NodeImpl<T> o2) {
				return comparator.compare(o1.getValue(), o2.getValue());
			}};
		sortNodes(root, cmp);
	}
	
	public static<T> void sortNodes(final NodeImpl<T> root, final Comparator<? super NodeImpl<T>> comparator) {
		Collections.sort(root.children, comparator);
		for (final NodeImpl<T> n : root.children) sortNodes(n, comparator);
	}

	public Node<E> getChild(int idx) {
		return getChildren().get(idx);
	}
	
	public Node<E> getParent(final int idx) {
		return getParents().get(idx);
	}
	
	public Node<E> getFirstParent() {
		final List<Node<E>> parents = getParents();
		if (parents.size() == 0) return null;
		return parents.get(0);
	}

	public static final Node<Object> EMPTY_ROOT = new Node<Object>(){

		@Override public Node<Object> get(final int idx) {
			return null;
		}

		@Override public Node<Object> get(final String id) {
			return null;
		}

		@Override public List<Node<Object>> getChildren() {
			return Collections.emptyList();
		}

		@Override public String getNodeId() {
			return "";
		}

		@Override public List<Node<Object>> getParents() {
			return Collections.emptyList();
		}

		@Override public Object getValue() {
			return null;
		}

		@Override public Node<Object> path(final String path) {
			return null;
		}

		@Override public List<String> getAllPaths() {
			return Collections.emptyList();
		}

		@Override public String getFirstPath() {
			return "";
		}

		@Override public boolean hasChildren() {
			return false;
		}

		@Override public boolean hasParents() {
			return false;
		}

		private static final long serialVersionUID = 1L;
	};
		
	@SuppressWarnings("unchecked")
	public static<E> Node<E> emptyRoot() {
		return (Node<E>) EMPTY_ROOT;
	}
	
	private static<E> void gatherPaths(
			final List<String> accumulator, 
			final StringBuilder currentPath, 
			final Node<E> node
	) {
		if (!node.hasParents()) {
			if (currentPath.length()>0) accumulator.add(currentPath.toString());
			return;
		}
		if (currentPath.length()>0) currentPath.insert(0, '/');
		currentPath.insert(0, node.getNodeId());
		for (final Node<E> c : node.getParents()) {
			gatherPaths(accumulator, new StringBuilder(currentPath), c);
		}
	}

	
	private static final long serialVersionUID = 1L;
}
