/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.collections;


/**
 * A typed version of org.hibernate.ScrollableResults
 * @author borisman
 *
 * @param <ENTITY>
 */
public interface Scrollable<ENTITY> {
	/**
	 * Advance to the next result
	 * @return <tt>true</tt> if there is another result
	 */
	public boolean next();
	/**
	 * Retreat to the previous result
	 * @return <tt>true</tt> if there is a previous result
	 */
	public boolean previous();
	/**
	 * Scroll an arbitrary number of locations
	 * @param i a positive (forward) or negative (backward) number of rows
	 * @return <tt>true</tt> if there is a result at the new location
	 */
	public boolean scroll(int i);
	/**
	 * Go to the last result
	 * @return <tt>true</tt> if there are any results
	 */
	public boolean last();
	/**
	 * Go to the first result
	 * @return <tt>true</tt> if there are any results
	 */
	public boolean first();
	/**
	 * Go to a location just before first result (this is the initial location)
	 */
	public void beforeFirst();
	/**
	 * Go to a location just after the last result
	 */
	public void afterLast();
	/**
	 * Is this the first result?
	 *
	 * @return <tt>true</tt> if this is the first row of results
	 * @throws HibernateException
	 */
	public boolean isFirst();
	/**
	 * Is this the last result?
	 *
	 * @return <tt>true</tt> if this is the last row of results
	 * @throws HibernateException
	 */
	public boolean isLast();
	/**
	 * Release resources immediately.
	 */
	public void close();
	/**
	 * Get currenct object
	 * @return an object or array
	 */
	public ENTITY get();
	/**
	 * Get the current location in the result set. The first
	 * row is number <tt>0</tt>, contrary to JDBC.
	 * @return the row number, numbered from <tt>0</tt>, or <tt>-1</tt> if
	 * there is no current row
	 */
	public int getRowNumber();
	/**
	 * Set the current location in the result set, numbered from either the
	 * first row (row number <tt>0</tt>), or the last row (row
	 * number <tt>-1</tt>).
	 * @param rowNumber the row number, numbered from the last row, in the
	 * case of a negative row number
	 * @return true if there is a row at that row number
	 */
	public boolean setRowNumber(int rowNumber);
	
}
