/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.collections.testlist;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

import org.builtu.exceptions.FatalError;
import org.builtu.testing.BasicTestCase;
import org.junit.Test;


/**
 * Basic induction testcase for List implementations. There are two types of tests: simple basic functionality 
 * assertions and random mutation testing. The {@link #testRandom_TesteeAgainstRef()} method performs 
 * comparing behavior of the "reference" implementation to actual "testee" implementation. "Reference" 
 * implementation is presumably correct. If "testee" behaves exactly as "reference" implementation, then "testee" 
 * is also correct.
 * 
 * By default, reference implementation is {@link java.util.ArrayList}, but can be redefined by overriding 
 * {@link #makeRef()} method.* 
 * 
 * The testcase performs random operations on both "reference" and "testee" implementation and observes the result. 
 * The "testee" implementation is correct, if for every operation:
 * <ol>
 * 	<li>
 * 		The exception class of the exception thrown from "reference" (if any) is 
 * 		equal to exception thrown by "testee". <br/>
 * 		Exceptions to this rule:
 * 		<ul>
 * 			<li>IndexOutOfBoundsException and ArrayIndexOutOfBoundsException considered the same</li>
 * 		</ul>
 * 	</li>
 * 	<li>
 * 		Results, returned from the random operation should be equal.
 * 	</li>
 * </ol>
 * 
 * This testcase is just a framework, so YOU should provide actual random operations for mutating lists. 
 * Additionally, there are simplier basic tests that are performed on both lists in order to check basic 
 * functionality.
 * 
 * BUGS: Currently sublists are not tested.
 * 
 * @author borisman
 */
public abstract class AbstractTestList<E> extends BasicTestCase {

	protected static interface ListOperation<E> {
		public E modifyList(final List<E> list);
	}

	protected E makeAnother(final E e) {
		E res = makeElement();
		for (int i = 0; i < 1000; i++) {
			if (!e.equals(res))
				return res;
			res = makeElement();
		}
		throw new FatalError("Can't make element that does not equals to " + e);
	}

	protected E makeAnother(final E e1, final E e2) {
		E res = makeElement();
		for (int i = 0; i < 1000; i++) {
			if (!e1.equals(res) && !e2.equals(res))
				return res;
			res = makeElement();
		}
		throw new FatalError("Can't make element that does not equals to " + e1
				+ " and " + e2);
	}

	protected abstract E makeElement();

	protected abstract ListOperation<E> makeModifier(final List<E> ref);

	protected List<E> makeRef() {
		final int size = rnd(200);
		final List<E> res = new ArrayList<E>();
		for (int i = 0; i < size; i++) {
			res.add(makeElement());
		}
		return res;
	}

	protected abstract List<E> makeTestee(final List<E> ref);

	@Test public void testRandom_TesteeAgainstRef() {
		final List<E> ref = makeRef();
		final List<E> tst = makeTestee(ref);
		assertEquals(ref, tst);

		for (int step = 0; step < 10000; step++) {
			final ListOperation<E> modifier = makeModifier(ref);
			boolean got_exception = false;
			E ref_result = null;
			try {
				ref_result = modifier.modifyList(ref);
			} catch (final Exception ref_exception) {
				got_exception = true;
				try {
					modifier.modifyList(tst);
					fail("step="+step+" Expected exception "+ref_exception);
				} catch (final Exception tst_exception) {
					final boolean ref_exception_outOfBounds = 
						ref_exception instanceof IndexOutOfBoundsException ||
						ref_exception instanceof ArrayIndexOutOfBoundsException;
					if (ref_exception_outOfBounds) {
						final boolean tst_exception_outOfBounds = 
							tst_exception instanceof IndexOutOfBoundsException || 
							tst_exception instanceof ArrayIndexOutOfBoundsException;
						assertTrue("step="+step+": reference="+ref_exception+", testee="+tst_exception,
								tst_exception_outOfBounds);
					} else {
						assertEquals("step="+step, ref_exception.getClass(), tst_exception.getClass());
					}
				}
			}
			if (!got_exception) {
				E tst_result = null;
				try {
					tst_result = modifier.modifyList(tst);
				} catch (final Exception e) {
					fail("step="+step+", got exception="+e);
				}
				if (ref_result == null && tst_result != null)
					fail("step="+step+" Expected null, got "+tst_result);
				assertEquals("step="+step, ref_result, tst_result);
			}
			this.assertEquals("step="+step, ref, tst);
		}
	}

	@Test public void testEquals() {
		final List<E> ref = makeRef();
		final List<E> tst = makeTestee(ref);
		
		assertEquals(ref, tst);
		assertEquals(tst, ref);
	}

	@Test public void testGet() {
		final List<E> ref = makeRef();
		final List<E> tst = makeTestee(ref);
		
		for (int i = 0; i < ref.size(); i++) {
			final int idx = rnd(ref.size());
			assertEquals(ref.get(idx), tst.get(idx));
		}
	}

	@Test public void testGetAtEnd() {
		final List<E> ref = makeRef();
		final List<E> tst = makeTestee(ref);
		
		try {
			tst.get(ref.size());
			fail();
		} 
		catch (final IndexOutOfBoundsException e) {}
	}

	@Test public void testGetBeforeStart() {
		final List<E> ref = makeRef();
		final List<E> tst = makeTestee(ref);
		
		try {
			tst.get(-1);
			fail();
		} catch (final IndexOutOfBoundsException e) {}
	}

	@Test public void testIterator() {
		final List<E> ref = makeRef();
		final List<E> tst = makeTestee(ref);
		
		final Iterator<E> refIt = ref.iterator();
		final Iterator<E> tstIt = tst.iterator();
		
		while (refIt.hasNext()) {
			assertTrue(tstIt.hasNext());
			assertEquals(refIt.next(), tstIt.next());
		}
		assertFalse(tstIt.hasNext());
	}

	@Test public void testListIterator() {
		final List<E> ref = makeRef();
		final List<E> tst = makeTestee(ref);
		
		final ListIterator<E> refIt = ref.listIterator();
		final ListIterator<E> tstIt = tst.listIterator();
		
		testListIterator(refIt, tstIt);
	}

	private void testListIterator(final ListIterator<E> refIt, final ListIterator<E> tstIt) {
		int idx = 0;
		while (refIt.hasNext()) {
			assertTrue(tstIt.hasNext());
			final int refIdx = refIt.nextIndex();
			final int tstIdx = tstIt.nextIndex();
			assertEquals("" + idx, refIdx, tstIdx);
			assertEquals("" + idx, refIt.next(), tstIt.next());
			idx++;
		}
		
		assertFalse(tstIt.hasNext());
		
		while (refIt.hasPrevious()) {
			assertTrue(tstIt.hasPrevious());
			assertEquals(refIt.previousIndex(), tstIt.previousIndex());
			assertEquals(refIt.previous(), tstIt.previous());
		}
	}

	@Test public void testListIteratorSkip() {
		final List<E> ref = makeRef();
		final List<E> tst = makeTestee(ref);
		
		if (ref.size() == 0) return;
		
		final int skip = rnd(ref.size());
		final ListIterator<E> refIt = ref.listIterator(skip);
		final ListIterator<E> tstIt = tst.listIterator(skip);
		
		testListIterator(refIt, tstIt);
	}

	@Test public void testListIteratorSkip0() {
		final List<E> ref = makeRef();
		final List<E> tst = makeTestee(ref);
		
		if (ref.size() == 0) return;
		
		final ListIterator<E> refIt = ref.listIterator(0);
		final ListIterator<E> tstIt = tst.listIterator(0);
		
		testListIterator(refIt, tstIt);
	}

	@Test public void testRangeExclusive() {
		final List<E> ref = makeRef();
		final List<E> tst = makeTestee(ref);
		
		testRangeExclusive(ref);
		testRangeExclusive(tst);
	}

	private void testRangeExclusive(final List<E> list) {
		if (list.size() == 0) return;
		assertEquals(list.size(), list.subList(0, list.size()).size());
	}

	@Test public void testSet() {
		final List<E> ref = makeRef();
		final List<E> tst = makeTestee(ref);
		
		for (int i = 0; i < ref.size(); i++) {
			final int idx = rnd(ref.size());
			final E e = makeElement();
			final E old = ref.set(idx, e);
			assertEquals(old, tst.set(idx, e));
			assertEquals(ref.get(idx), tst.get(idx));
		}
	}

	@Test public void testSize() {
		final List<E> ref = makeRef();
		final List<E> tst = makeTestee(ref);
		
		assertEquals(ref.size(), tst.size());
	}
}
