/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.config.module;

import java.io.File;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.apache.commons.configuration.CombinedConfiguration;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.DefaultConfigurationBuilder;
import org.apache.commons.configuration.FileConfiguration;
import org.apache.commons.configuration.reloading.ManagedReloadingStrategy;
import org.builtu.exceptions.FatalError;
import org.builtu.exceptions.ResourceNotFoundException;
import org.builtu.util.commutil.StringUtil;


/**
 * Provides conveinient methods for application site-specific configuration.
 * My vision on site-specific configuration is following:
 * <h2>Axioms</h2>
 * <ol>
 *     <li> 
 *         In multiproject application, where one MAIN project is linked against several subprojects...
 *         <ol>
 *             <li>...no subproject should have it's own site-specific configuration ...</li>
 *             <li>...all site-specific configuration should be done by MAIN project.</li>
 *         </ol>
 *     </li>
 *     <li> 
 *         Application can have several "layers" of configuration:
 *         <ol>
 *         		<li> compile-time configuration (hardcoded into binary), </li>
 *              <li> site-wide configuration, </li>
 *              <li> user configuration. </li>
 *         </ol>
 *         More "layers" can be added or some can be removed on application-per-application basis according to 
 *         requirements. 
 *     </li>
 *     <li> Each successive configuration layer (normally) overrides values from previous layers. </li>
 *     <li> 
 *     		In best UNIX traditions, 
 *     		<ol>
 *     			<li>site-wide configuration is stored in /etc/myapp/ folder,</li>
 *     			<li>user-specific configuration is stored in $HOME/.myapp/ folder</li>
 *     		</ol>
 *     </li> 
 *     <li> 
 *     		To simplify matters, configuration files have same names as applications, for example, ws_domdot application
 *     		should have it's configuration in /etc/yellteam/myapp.xml and $HOME/.yellteam/myapp.xml
 *     </li>
 * </ol>
 * 
 * In order to implement those axioms, Commons Configuraton is used. You normally put a file called myapp.xml at the
 * root of your .jar or .war so it will be accessible at "/myapp.xml" classpath.
 * Typical myapp.xml should look like:
 * <pre>
 * &lt;?xml version="1.0" encoding="UTF-8" ?&gt;
 * 
 * &lt;configuration&gt;
 *     &lt;override&gt;
 *     		&lt;xml fileName="${sys:user.home}/.yellteam/myapp.xml" optional="true"/&gt;
 *     		&lt;xml fileName="etc/yellteam/myapp.xml" optional="true"/&gt;
 *          &lt;xml fileName="myapp_default.xml" /&gt;
 *     &lt;/override&gt;
 * &lt;/configuration>
 * </pre>
 * 
 * In this case, config_default.xml will be used for compile-time configuration, etc/yellteam/myapp.xml will be used 
 * for site-wide configuration, ${sys:user.home} will get expanded to current user home directory at runtime and 
 * ${sys:user.home}/.yellteam/myapp.xml will point to user-sepecific configuration.
 * 
 * Files that are included first override files included later.
 * Now you can create myapp_default.xml:
 * <pre>
 * &lt;configuration&gt;
 *     &lt;myapp&gt;
 *         &lt;name&gt;
 *             value
 *         &lt;/name&gt;
 *     &lt;/myapp&gt;
 * &lt;/configuration&gt;
 * </pre>
 * This is simple xml format Commons Configuration use (you can use any other format as well, but xml is one of the 
 * easiest and localization-friendly). In the example above, value can be retrieved later by calling:
 * <code>
 * final Config config = new Config.get("myapp");
 * final String value = config.getString("myapp.name");
 * </code>   
 * @author borisman
 *
 */
public class Config {
	
	private final CombinedConfiguration conf;
	private final ManagedReloadingStrategy strategy = new ManagedReloadingStrategy();
	private final File lastConfig;
	private final String [] configFiles;
	
	public Config(final String classpath) {
		URL configURL = Config.class.getResource(classpath);
		if (configURL == null && !classpath.startsWith("/")) configURL = Config.class.getResource('/'+classpath);
		if (configURL == null) throw new ResourceNotFoundException(classpath);
    	try {
    		final DefaultConfigurationBuilder builder = new DefaultConfigurationBuilder(configURL);
    		builder.setReloadingStrategy(strategy);
    		this.conf = builder.getConfiguration(true);
   			this.lastConfig = resolveLastModifiableConfig(conf);
    		this.configFiles = dumpConfigFiles(conf); 
		} catch (final ConfigurationException e) {
			throw new FatalError(e);
		}
	}
	
	private String[] dumpConfigFiles(final CombinedConfiguration conf) {
		final List<String> res = new ArrayList<String>();
		for (int i=conf.getNumberOfConfigurations()-1; i>=0; i--) {
			final Configuration c = conf.getConfiguration(i);
			if (!(c instanceof FileConfiguration)) continue;
			final FileConfiguration fc = (FileConfiguration) c;
			final File f = fc.getFile();
			if (f == null) continue;
			final String file = f.getAbsolutePath();
			if (file == null) continue;
			res.add(file);
		}
		final String [] arr = new String [res.size()];
		res.toArray(arr);
		return arr;
	}

	private static File resolveLastModifiableConfig(final CombinedConfiguration conf) {
		if (conf.getNumberOfConfigurations()<=0) return null; 
		final Configuration last = conf.getConfiguration(0);
		if (last instanceof FileConfiguration) {
			return ((FileConfiguration) last).getFile(); 
		} else { 
			return null;
		}
	}

	public void reload() {
		strategy.refresh();
	}

	public String getString(final String key) throws ConfigParameterNotFoundError {
		final List<?> res = conf.getList(key);
    	if (res == null) {
    		throw new ConfigParameterNotFoundError(configFiles, key);
    	}
    	return StringUtil.join(',', res);
    }

	/*public String getString(final String key, final String defaultVal) {
    	return conf.getString(key, defaultVal);
    }*/
	
	public void setString(final String key, final String value) {
		conf.setProperty(key, value);
	}
	
	public int getInt(final String key) throws ConfigParameterNotFoundError {
    	if (!conf.containsKey(key)) {
    		throw new ConfigParameterNotFoundError(configFiles, key);
    	}
    	return conf.getInt(key);
	}

	public long getLong(final String key) {
    	if (!conf.containsKey(key)) {
    		throw new ConfigParameterNotFoundError(configFiles, key);
    	}
    	return conf.getLong(key);
	}
	
	public boolean getBoolean(final String key) throws ConfigParameterNotFoundError {
    	if (!conf.containsKey(key)) {
    		throw new ConfigParameterNotFoundError(configFiles, key);
    	}
    	return conf.getBoolean(key);
	}
	
	public void setBoolean(final String key, final boolean value) {
		conf.setProperty(key, value);
	}

	/*public int getInt(final String key, final int devaultVal) {
    	return conf.getInt(key, devaultVal);
	}*/
	
	public void setInt(final String key, final int value) throws ConfigParameterNotFoundError {
		conf.setProperty(key, value);
	}
    
    public Properties getProperties(final String key) throws ConfigParameterNotFoundError {
    	if (!conf.containsKey(key)) {
    		throw new ConfigParameterNotFoundError(configFiles, key);
    	}
    	return conf.getProperties(key);
    }
	
	public Iterable<String> getKeys() {
    	return new Iterable<String> () {
			@Override public Iterator<String> iterator() {
				return getKeysIterator();
			}
    	};
    }

	@SuppressWarnings("unchecked")
	public Iterator<String> getKeysIterator() {
		return (Iterator<String>) conf.getKeys();
	}

	public Iterable<String> getKeys(final String prefix) {
    	return new Iterable<String> () {
			@Override public Iterator<String> iterator() {
				return getKeysIterator(prefix);
			}
    	};
    }
	
    @SuppressWarnings("unchecked")
	public Iterator<String> getKeysIterator(final String prefix) {
    	return conf.getKeys(prefix);
    }
    
    public File getLastConfigurationFile()  {
    	return this.lastConfig;
    }
    
	public String dumpKeyAsXml(final String key) {
		final StringBuilder res = new StringBuilder(2048);
		res.append(XML_TOP);
		final String [] keypath = key.split("\\.");
		writeKeyProlog(res, keypath);
		dumpKey(keypath.length, res, key);
		writekeyEpilog(res, keypath);
		res.append("</configuration>");
		return res.toString();
	}
    
	private void dumpKey(final int depth, final StringBuilder res, final String key) {
		final String contents = conf.getString(key);
		if (contents != null) {
			pad(res, depth+1);
			res.append(contents);
			res.append('\n');
			return;
		}
		final Iterator<String> subkeys = getKeysIterator(key);
		while (subkeys.hasNext()) {
			final String subkey = subkeys.next();
			assert subkey.startsWith(key+'.');
			final String name = subkey.substring(key.length()+1);
			pad(res, depth);
			res.append('<');
			res.append(name);
			res.append(">\n");
			dumpKey(depth+1, res, subkey);
			pad(res, depth);
			res.append("</");
			res.append(name);
			res.append(">\n");
		}
		
	}

	private void writeKeyProlog(final StringBuilder res, final String[] keypath) {
		for (int i=0; i<keypath.length; i++) {
			pad(res, i);
			res.append('<');
			res.append(keypath[i]);
			res.append(">\n");
		}
	}

	private void writekeyEpilog(final StringBuilder res, final String[] keypath) {
		for (int i=keypath.length-1; i>=0; i--) {
			pad(res, i);
			res.append("</");
			res.append(keypath[i]);
			res.append(">\n");
		}
	}
	
	private void pad(final StringBuilder b, final int nowmuch) {
		for (int padding=0; padding<nowmuch; padding++) b.append('\t');
	}

	private static final String XML_TOP="<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n<configuration>\n";
}
