/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.persistence.hibernate;

import java.io.Serializable;
import java.util.List;
import java.util.regex.Pattern;

import org.builtu.assertlevel.AL;
import org.builtu.persistence.DAO;
import org.builtu.persistence.OrderColumn;
import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.Session;
import org.springframework.orm.hibernate3.HibernateCallback;
import org.springframework.orm.hibernate3.HibernateTemplate;
import org.springframework.orm.hibernate3.SessionFactoryUtils;
import org.springframework.orm.hibernate3.support.HibernateDaoSupport;


public class DAOImpl<PK extends Serializable, ENTITY> extends HibernateDaoSupport implements DAO<PK, ENTITY> {
	
	private final Class<ENTITY> entityClass;
	
	protected DAOImpl (final Class<ENTITY> entityClass) {
		this.entityClass = entityClass;
	}
	
	@Override public ENTITY get(final PK id) {
		return (ENTITY) super.getHibernateTemplate().get(entityClass, id);
	}

	@Override public void save(final ENTITY entity) {
		super.getHibernateTemplate().saveOrUpdate(entity);
	}

	@SuppressWarnings("unchecked")
	protected ENTITY find(final String hql, final Object ... params) {
		checkNonNull(params);
		final List<?> res = super.getHibernateTemplate().find(hql, params);
		if (res.size()==0) return null;
		if (res.size()>1) throw new RuntimeException("Multiple values found for: "+hql+", params="+params);
		
		return (ENTITY) res.get(0);
	}

        @Override
        public void delete(ENTITY entity) {
            super.getHibernateTemplate().delete(entity);
        }


	/**
	 * @param orderBy when not null, specifies ordering by one or more columns. Possible values for columns are:
	 *                This could be used with asc/desc specification ("date desc", "transactionId asc") 
	 * @param offset when not null, specifies from where results should be returned
	 * @param max when not null, specifies how many results should be returned	 
	 */
	protected List<ENTITY> list(
			final String hql, 
			final Integer offset, 
			final Integer max, 
			final OrderColumn [] orderBy,
			final Object ... params
	) {
		checkNonNull(params);
		checkNonNull(orderBy);
		assert AL.input() || !hqlHasOrderBy(hql);
		
		final StringBuilder order = new StringBuilder(params.length << 4);
		
		if (orderBy != null) {
			for (final OrderColumn column : orderBy) {
				if (order.length()!=0) order.append(',');
				order.append(column.toString());
			}
		}
		
		final StringBuilder fullQuery = new StringBuilder(order.length()+hql.length());
		fullQuery.append(hql);
		if (order.length()>0) {
			fullQuery.append(" ORDER BY ");
			fullQuery.append(order);
		}
		
		return super.getHibernateTemplate().executeWithNativeSession(new HibernateCallback<List<ENTITY>>() {
			@SuppressWarnings("unchecked")
			public List<ENTITY> doInHibernate(Session session) throws HibernateException {
				Query queryObject = session.createQuery(fullQuery.toString());
				prepareQuery(queryObject);
				if (params != null) {
					for (int i = 0; i < params.length; i++) {
						queryObject.setParameter(i, params[i]);
					}
				}
				if (offset!=null) queryObject.setFirstResult(offset);
				if (max!=null) queryObject.setMaxResults(max);

				return queryObject.list();
			}
		});
	}
	
	private static final Pattern ORDER_BY = Pattern.compile("ORDER\\s+BY", Pattern.CASE_INSENSITIVE);
	
	private boolean hqlHasOrderBy(String hql) {
		return ORDER_BY.matcher(hql).find();
	}

	/**
	 * Got from from {@link HibernateTemplate#prepareQuery(Query)}
	 * Prepare the given Query object, applying cache settings and/or
	 * a transaction timeout.
	 * @param queryObject the Query object to prepare
	 * @see #setCacheQueries
	 * @see #setQueryCacheRegion
	 * @see SessionFactoryUtils#applyTransactionTimeout
	 */
	protected void prepareQuery(Query queryObject) {
		final HibernateTemplate tpl = super.getHibernateTemplate();
		if (tpl.isCacheQueries()) {
			queryObject.setCacheable(true);
			if (tpl.getQueryCacheRegion() != null) {
				queryObject.setCacheRegion(tpl.getQueryCacheRegion());
			}
		}
		if (tpl.getFetchSize() > 0) {
			queryObject.setFetchSize(tpl.getFetchSize());
		}
		if (tpl.getMaxResults() > 0) {
			queryObject.setMaxResults(tpl.getMaxResults());
		}
		SessionFactoryUtils.applyTransactionTimeout(queryObject, getSessionFactory());
	}
	

	private void checkNonNull(final Object [] params) {
		for (final Object o : params) {
			if (o == null) throw new RuntimeException("Can't find using null parameter in: "+params);
		}
	}
}
