/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.services.resources;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.builtu.config.module.Config;
import org.builtu.config.module.ConfigUtil;
import org.builtu.config.module.Configurable;
import org.builtu.util.commutil.StringUtil;


public class ResourceRoot implements Configurable {
	
	public static final String MODULE = "static-resources";
	
	private List<ResourceLocation> locations;
	
	/**
	 * Skip XML configuration and provide search paths as array
	 * @param searchPaths
	 */
	public ResourceRoot(final String ... searchPaths) {
		final List<String> paths = new ArrayList<String>(searchPaths.length);
		for (final String s : searchPaths) {
			paths.add(s);
		}
		setLocations(paths);
	}

	/**
	 * Skip XML configuration and provide search paths as list
	 * @param searchPaths
	 */
	public ResourceRoot(final List<String> searchPaths) {
		setLocations(searchPaths);
	}
	
	/**
	 * Configure ResourceRoot via xml config
	 */
	public ResourceRoot() {
		ConfigUtil.configure(MODULE, this);
	}

	@Override public void configure(final String moduleName, final Config config) {
		assert moduleName == MODULE;
		final String paths = config.getString(MODULE+".paths");
		setLocations(StringUtil.split_list(paths, '\n'));
	}
	
	private void setLocations(final List<String> paths) {
		final String CLASSPATH="classpath://"; 
		final String FILE="file://";
		
		locations = new ArrayList<ResourceLocation>(paths.size());
		for (String path : paths) {
			path = path.trim();
			path = removeTrailingSlashes(path);
			
			final ResourceLocation location;
			if (path.startsWith(CLASSPATH)) {
				location = new ClasspathResourceLocation(path.substring(CLASSPATH.length()));
			} else if (path.startsWith(FILE)) {
				location = new FileResourceLocation(new File(path.substring(FILE.length())));
			} else {
				location = new UrlResourceLocation (path);
			}
			
			locations.add(location);
		}
	}
	
	private static String removeTrailingSlashes(final String path) {
		int i;
		for (i = path.length()-1; i>=0; i--) {
			if (path.charAt(i) != '/' && path.charAt(i) != '\\') break;
		}
		return path.substring(0, i+1);
	}

	public Resource get(final String path) {
		if (path.length() == 0 || path.charAt(0) != '/') {
			throw new IllegalArgumentException("path should start with '/', i got instead: '"+path+'\'');
		}
		for (final ResourceLocation location : locations) {
			final Resource res = location.resolve(path);
			if (res != null) return res;
		}
		return null;
	}

}
