/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.util.commutil;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

import junit.framework.Assert;

import org.builtu.testing.BasicTestCase;
import org.builtu.util.commutil.IOUtil;
import org.junit.Test;


public class TestIOUtil extends BasicTestCase {
	
	private static final String UTF8MESSAGE="привет hello";
	private static final String ASCIIMESSAGE="hello";

	@Test public void test_writeRaw() throws IOException {
		final File test = super.createTempFile();
		
		IOUtil.writeRaw(test, UTF8MESSAGE.getBytes("UTF-8"));
		
		final String read;
		final InputStream in = new FileInputStream(test);
		try {
			Assert.assertTrue(test.length()<=8*UTF8MESSAGE.length());
			final int length = (int) test.length();
			final byte [] data = new byte [length];
			in.read(data);
			read = new String(data, "UTF-8");
		} finally {
			in.close();
		}
		
		Assert.assertEquals(UTF8MESSAGE, read);
		
	}

	@Test public void test_readRaw() throws IOException {
		final File test = super.createTempFile();
		
		IOUtil.writeUTF8FileContents(test, UTF8MESSAGE);
		
		final byte [] data = IOUtil.readRaw(test);
		
		super.assertArraysEquals(UTF8MESSAGE.getBytes("UTF-8"), data);
		
	}
	
	@Test public void test_writeUTF8FileContents() throws IOException {
		final File test = super.createTempFile();
		
		final OutputStream out = new FileOutputStream(test);
		try {
			out.write(UTF8MESSAGE.getBytes("UTF-8"));
		} finally {
			out.close();
		}

		final String read;
		final InputStream in = new FileInputStream(test);
		try {
			Assert.assertTrue(test.length()<=8*UTF8MESSAGE.length());
			final int length = (int) test.length();
			final byte [] data = new byte [length];
			in.read(data);
			read = new String(data, "UTF-8");
		} finally {
			in.close();
		}
		
		Assert.assertEquals(UTF8MESSAGE, read);
	}

	@Test public void test_readUTF8FileContents() throws IOException {
		final File test = super.createTempFile();
		final OutputStream out = new FileOutputStream(test);
		try {
			out.write(UTF8MESSAGE.getBytes("UTF-8"));
		} finally {
			out.close();
		}

		final String read = IOUtil.readUTF8FileContents(test);
		Assert.assertEquals(UTF8MESSAGE, read);
	}
	
	@Test public void test_writeASCIIFileContents() throws IOException {
		final File test = super.createTempFile();
		
		IOUtil.writeASCIIFileContents(test, ASCIIMESSAGE);
		
		final String read;
		final InputStream in = new FileInputStream(test);
		try {
			Assert.assertTrue(test.length()<=8*ASCIIMESSAGE.length());
			final int length = (int) test.length();
			final byte [] data = new byte [length];
			in.read(data);
			read = new String(data, "UTF-8");
		} finally {
			in.close();
		}
		
		Assert.assertEquals(ASCIIMESSAGE, read);
	}

	@Test public void test_readASCIIFileContents() throws IOException {
		final File test = super.createTempFile();
		final OutputStream out = new FileOutputStream(test);
		try {
			out.write(ASCIIMESSAGE.getBytes("UTF-8"));
		} finally {
			out.close();
		}
		
		final String read = IOUtil.readASCIIFileContents(test);
		Assert.assertEquals(ASCIIMESSAGE, read);
	}
	
	@Test public void test_UTF8_RW() throws IOException {
		final File test = super.createTempFile();
		IOUtil.writeUTF8FileContents(test, UTF8MESSAGE);
		final String read = IOUtil.readUTF8FileContents(test);
		Assert.assertEquals(UTF8MESSAGE, read);
	}
	
	@Test public void test_ASCII_RW() throws IOException {
		final File test = super.createTempFile();
		IOUtil.writeASCIIFileContents(test, ASCIIMESSAGE);
		final String read = IOUtil.readASCIIFileContents(test);
		Assert.assertEquals(ASCIIMESSAGE, read);
	}

	@Test public void test_ASCII_UTF8_RW() throws IOException {
		final File test = super.createTempFile();
		IOUtil.writeASCIIFileContents(test, ASCIIMESSAGE);
		final String read = IOUtil.readUTF8FileContents(test);
		Assert.assertEquals(ASCIIMESSAGE, read);
	}

	@Test public void test_UTF8_ASCII_RW() throws IOException {
		final File test = super.createTempFile();
		IOUtil.writeUTF8FileContents(test, ASCIIMESSAGE);
		final String read = IOUtil.readASCIIFileContents(test);
		Assert.assertEquals(ASCIIMESSAGE, read);
	}
	
	@Test public void testStream () throws IOException {
		final File inf = super.createTempFile();
		final File outf = super.createTempFile();
		IOUtil.writeUTF8FileContents(inf, UTF8MESSAGE);
		final InputStream in = new FileInputStream(inf);
		try {
			final OutputStream out = new FileOutputStream(outf);
			try {
				IOUtil.stream(in, out);
			} finally {
				out.close();
			}
		} finally {
			in.close();
		}
		Assert.assertEquals(UTF8MESSAGE, IOUtil.readUTF8FileContents(outf));
	}
	
	@Test public void test_readASCIIStrings() throws IOException {
		final File test = super.createTempFile();
		
		final String lines = "line1\n line2 \nline3";
		IOUtil.writeASCIIFileContents(test, lines);
		
		final List<String> read = IOUtil.readASCIIStrings(test);
		
		Assert.assertEquals(3, read.size());
		Assert.assertEquals("line1", read.get(0));
		Assert.assertEquals(" line2 ", read.get(1));
		Assert.assertEquals("line3", read.get(2));
	}
	
	@Test public void test_writeASCIIStrings() throws IOException {
		final File test = super.createTempFile();
		
		final List<String> strings = new ArrayList<String>(3);
		strings.add("line1");
		strings.add("line2");
		strings.add("line3");
		
		IOUtil.writeASCIIStrings(test, strings);
		
		Assert.assertEquals("line1\nline2\nline3", IOUtil.readASCIIFileContents(test));
	}

	@Test public void test_readUTF8Strings() throws IOException {
		final File test = super.createTempFile();
		
		final String lines = "строка line1\n строка line2 \nстрока line3";
		IOUtil.writeUTF8FileContents(test, lines);
		
		final List<String> read = IOUtil.readUTF8Strings(test);
		
		Assert.assertEquals(3, read.size());
		Assert.assertEquals("строка line1", read.get(0));
		Assert.assertEquals(" строка line2 ", read.get(1));
		Assert.assertEquals("строка line3", read.get(2));
	}
	
	@Test public void test_writeUTF8Strings() throws IOException {
		final File test = super.createTempFile();
		
		final List<String> strings = new ArrayList<String>(3);
		strings.add("строка line1");
		strings.add("строка line2");
		strings.add("строка line3");
		
		IOUtil.writeUTF8Strings(test, strings);
		
		Assert.assertEquals("строка line1\nстрока line2\nстрока line3", IOUtil.readUTF8FileContents(test));
	}

	private static final List<String> ASCII_STRINGS;
	static {
		ASCII_STRINGS = new ArrayList<String>(3);
		ASCII_STRINGS.add("line1");
		ASCII_STRINGS.add("line2");
		ASCII_STRINGS.add("line3");
	}
	
	@Test public void test_ASCIIStrings_RW() throws IOException {
		final File test = super.createTempFile();
		
		IOUtil.writeASCIIStrings(test, ASCII_STRINGS);
		
		Assert.assertEquals(ASCII_STRINGS, IOUtil.readASCIIStrings(test));
	}
	
	@Test public void test_writeASCIIStrings_doesnotfailWhenStringsContainCRLF() throws IOException {
		final File test = super.createTempFile();
		
		final List<String> sample = new ArrayList<String>(3);
		sample.add("line1\nline2");
		sample.add("line3");
		
		IOUtil.writeASCIIStrings(test, sample);
		
		final List<String> read = IOUtil.readASCIIStrings(test);
		
		Assert.assertEquals(3, read.size());
		Assert.assertEquals("line1", read.get(0));
		Assert.assertEquals("line2", read.get(1));
		Assert.assertEquals("line3", read.get(2));
	}
	
	private static final List<String> UTF8_STRINGS;
	static {
		UTF8_STRINGS = new ArrayList<String>(3);
		UTF8_STRINGS.add("строка line1");
		UTF8_STRINGS.add("строка line2");
		UTF8_STRINGS.add("строка line3");
	}
	
	@Test public void test_UTF8Strings_RW() throws IOException {
		final File test = super.createTempFile();
		
		IOUtil.writeUTF8Strings(test, UTF8_STRINGS);
		
		Assert.assertEquals(UTF8_STRINGS, IOUtil.readUTF8Strings(test));
	}
	
	@Test public void test_writeUTF8Strings_doesnotfailWhenStringsContainCRLF() throws IOException {
		final File test = super.createTempFile();
		
		final List<String> sample = new ArrayList<String>(3);
		sample.add("строка line1\nстрока line2");
		sample.add("строка line3");
		
		IOUtil.writeUTF8Strings(test, sample);
		
		final List<String> read = IOUtil.readUTF8Strings(test);
		
		Assert.assertEquals(3, read.size());
		Assert.assertEquals("строка line1", read.get(0));
		Assert.assertEquals("строка line2", read.get(1));
		Assert.assertEquals("строка line3", read.get(2));
	}
	

	
}
