/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.util.html;

import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class HtmlUtil {
    private static final Pattern html_start = Pattern.compile("<html.*?>", Pattern.CASE_INSENSITIVE);
    private static final Pattern body_start = Pattern.compile("<body.*?>", Pattern.CASE_INSENSITIVE);
    private static final Pattern html_end = Pattern.compile("</html.*?>", Pattern.CASE_INSENSITIVE);
    private static final Pattern body_end = Pattern.compile("</body.*?>", Pattern.CASE_INSENSITIVE);

    private static final Map<Pattern, String> HTML2TEXT = new LinkedHashMap<Pattern, String>();

    static {
        //remove line brakes
        HTML2TEXT.put(Pattern.compile("\\r\\n|\\r|\\n|\\t"), " ");
        //remove HTML comments
        HTML2TEXT.put(Pattern.compile("<!--.*?-->"), " ");
        //remove other HTML tags
        HTML2TEXT.put(Pattern.compile("<[^>]+>"), " ");

        //replcase special characters
        //see http://www.theukwebdesigncompany.com/articles/entity-escape-characters.php
        HTML2TEXT.put(Pattern.compile("&euro;"), "€");
        HTML2TEXT.put(Pattern.compile("&#32;"), " ");
        HTML2TEXT.put(Pattern.compile("&#33;"), "!");
        HTML2TEXT.put(Pattern.compile("&#34;"), "\"");
        HTML2TEXT.put(Pattern.compile("&quot;"), "\"");
        HTML2TEXT.put(Pattern.compile("&#35;"), "#");
        HTML2TEXT.put(Pattern.compile("&#36;"), "$");
        HTML2TEXT.put(Pattern.compile("&#37;"), "%");
        HTML2TEXT.put(Pattern.compile("&#39;"), "'");
        HTML2TEXT.put(Pattern.compile("&#40;"), "(");
        HTML2TEXT.put(Pattern.compile("&#41;"), ")");
        HTML2TEXT.put(Pattern.compile("&#42;"), "*");


        HTML2TEXT.put(Pattern.compile("&#43;"), "+");
        HTML2TEXT.put(Pattern.compile("&#44;"), ",");
        HTML2TEXT.put(Pattern.compile("&#45;"), "-");
        HTML2TEXT.put(Pattern.compile("&#46;"), ".");
        HTML2TEXT.put(Pattern.compile("&#47;"), "/");
        HTML2TEXT.put(Pattern.compile("&#48;"), "0");
        HTML2TEXT.put(Pattern.compile("&#49;"), "1");
        HTML2TEXT.put(Pattern.compile("&#50;"), "2");
        HTML2TEXT.put(Pattern.compile("&#51;"), "3");
        HTML2TEXT.put(Pattern.compile("&#52;"), "4");
        HTML2TEXT.put(Pattern.compile("&#53;"), "5");
        HTML2TEXT.put(Pattern.compile("&#54;"), "6");
        HTML2TEXT.put(Pattern.compile("&#55;"), "7");
        HTML2TEXT.put(Pattern.compile("&#56;"), "8");
        HTML2TEXT.put(Pattern.compile("&#57;"), "9");
        HTML2TEXT.put(Pattern.compile("&#58"), ":");
        HTML2TEXT.put(Pattern.compile("&#59;"), ";");
        HTML2TEXT.put(Pattern.compile("&#60;"), "<");
        HTML2TEXT.put(Pattern.compile("&lt;"), "<");
        HTML2TEXT.put(Pattern.compile("&#61;"), "=");
        HTML2TEXT.put(Pattern.compile("&#62;"), ">");
        HTML2TEXT.put(Pattern.compile("&gt;"), ">");
        HTML2TEXT.put(Pattern.compile("&#63;"), "?");
        HTML2TEXT.put(Pattern.compile("&#64;"), "@");
/*  -- Commented by RS to minimize quantity of actual Patterns
        HTML2TEXT.put(Pattern.compile("&#65;"), "A");
        HTML2TEXT.put(Pattern.compile("&#66;"), "B");
        HTML2TEXT.put(Pattern.compile("&#67;"), "C");
        HTML2TEXT.put(Pattern.compile("&#68;"), "D");
        HTML2TEXT.put(Pattern.compile("&#69;"), "E");
        HTML2TEXT.put(Pattern.compile("&#70;"), "F");
        HTML2TEXT.put(Pattern.compile("&#71;"), "G");
        HTML2TEXT.put(Pattern.compile("&#72;"), "H");
        HTML2TEXT.put(Pattern.compile("&#73;"), "I");
        HTML2TEXT.put(Pattern.compile("&#74;"), "J");
        HTML2TEXT.put(Pattern.compile("&#75;"), "K");
        HTML2TEXT.put(Pattern.compile("&#76;"), "L");
        HTML2TEXT.put(Pattern.compile("&#77;"), "M");
        HTML2TEXT.put(Pattern.compile("&#78;"), "N");
        HTML2TEXT.put(Pattern.compile("&#79;"), "O");
        HTML2TEXT.put(Pattern.compile("&#80;"), "P");
        HTML2TEXT.put(Pattern.compile("&#81;"), "Q");
        HTML2TEXT.put(Pattern.compile("&#82;"), "R");
        HTML2TEXT.put(Pattern.compile("&#83;"), "S");
        HTML2TEXT.put(Pattern.compile("&#84;"), "T");
        HTML2TEXT.put(Pattern.compile("&#85;"), "U");
        HTML2TEXT.put(Pattern.compile("&#86;"), "V");
        HTML2TEXT.put(Pattern.compile("&#87;"), "W");
        HTML2TEXT.put(Pattern.compile("&#88;"), "X");
        HTML2TEXT.put(Pattern.compile("&#89;"), "Y");
        HTML2TEXT.put(Pattern.compile("&#90;"), "Z");
*/
        HTML2TEXT.put(Pattern.compile("&#91;"), "[");
        HTML2TEXT.put(Pattern.compile("&#92;"), "\\");
        HTML2TEXT.put(Pattern.compile("&#93;"), "]");
        HTML2TEXT.put(Pattern.compile("&#94;"), "^");
        HTML2TEXT.put(Pattern.compile("&#95;"), "_");
        HTML2TEXT.put(Pattern.compile("&#96;"), "`");
/*  -- Commented by RS to minimize quantity of actual Patterns
        HTML2TEXT.put(Pattern.compile("&#97;"), "a");
        HTML2TEXT.put(Pattern.compile("&#98;"), "b");
        HTML2TEXT.put(Pattern.compile("&#99;"), "c");
        HTML2TEXT.put(Pattern.compile("&#100;"), "d");
        HTML2TEXT.put(Pattern.compile("&#101;"), "e");
        HTML2TEXT.put(Pattern.compile("&#102;"), "f");
        HTML2TEXT.put(Pattern.compile("&#103;"), "g");
        HTML2TEXT.put(Pattern.compile("&#104;"), "h");
        HTML2TEXT.put(Pattern.compile("&#105;"), "i");
        HTML2TEXT.put(Pattern.compile("&#106;"), "j");
        HTML2TEXT.put(Pattern.compile("&#107;"), "k");
        HTML2TEXT.put(Pattern.compile("&#108;"), "l");
        HTML2TEXT.put(Pattern.compile("&#109;"), "m");
        HTML2TEXT.put(Pattern.compile("&#110;"), "n");
        HTML2TEXT.put(Pattern.compile("&#111;"), "o");
        HTML2TEXT.put(Pattern.compile("&#112;"), "p");
        HTML2TEXT.put(Pattern.compile("&#113;"), "q");
        HTML2TEXT.put(Pattern.compile("&#114;"), "r");
        HTML2TEXT.put(Pattern.compile("&#115;"), "s");
        HTML2TEXT.put(Pattern.compile("&#116;"), "t");
        HTML2TEXT.put(Pattern.compile("&#117;"), "u");
        HTML2TEXT.put(Pattern.compile("&#118;"), "v");
        HTML2TEXT.put(Pattern.compile("&#119;"), "w");
        HTML2TEXT.put(Pattern.compile("&#120;"), "x");
        HTML2TEXT.put(Pattern.compile("&#121;"), "y");
        HTML2TEXT.put(Pattern.compile("&#122;"), "z");
*/
        HTML2TEXT.put(Pattern.compile("&#123;"), "{");
        HTML2TEXT.put(Pattern.compile("&#124;"), "|");
        HTML2TEXT.put(Pattern.compile("&#125;"), "}");
        HTML2TEXT.put(Pattern.compile("&#126;"), "~");
        HTML2TEXT.put(Pattern.compile("&#160;"), " ");
        HTML2TEXT.put(Pattern.compile("&nbsp;"), " ");
        HTML2TEXT.put(Pattern.compile("&#161;"), "¡");
        HTML2TEXT.put(Pattern.compile("&iexcl;"), "¡");
        HTML2TEXT.put(Pattern.compile("&#162;"), "¢");
        HTML2TEXT.put(Pattern.compile("&cent;"), "¢");
        HTML2TEXT.put(Pattern.compile("&#163;"), "£");
        HTML2TEXT.put(Pattern.compile("&pound;"), "£");
        HTML2TEXT.put(Pattern.compile("&#164;"), "¤");
        HTML2TEXT.put(Pattern.compile("&curren;"), "¤");
        HTML2TEXT.put(Pattern.compile("&#165;"), "¥");
        HTML2TEXT.put(Pattern.compile("&yen;"), "¥");
        HTML2TEXT.put(Pattern.compile("&#166;"), "¦");
        HTML2TEXT.put(Pattern.compile("&brvbar;"), "¦");
        HTML2TEXT.put(Pattern.compile("&#167;"), "§");
        HTML2TEXT.put(Pattern.compile("&sect;"), "§");
        HTML2TEXT.put(Pattern.compile("&#168;"), "¨");
        HTML2TEXT.put(Pattern.compile("&uml;"), "¨");
        HTML2TEXT.put(Pattern.compile("&#169;"), "©");
        HTML2TEXT.put(Pattern.compile("&copy;"), "");
        HTML2TEXT.put(Pattern.compile("&#170;"), "ª");
        HTML2TEXT.put(Pattern.compile("&ordf;"), "ª");
        HTML2TEXT.put(Pattern.compile("&#171;"), "«");
        HTML2TEXT.put(Pattern.compile("&#172;"), "¬");
        HTML2TEXT.put(Pattern.compile("&not;"), "¬");
        HTML2TEXT.put(Pattern.compile("&#173;"), "­");
        HTML2TEXT.put(Pattern.compile("&shy;"), "");
        HTML2TEXT.put(Pattern.compile("&#174;"), "®");
        HTML2TEXT.put(Pattern.compile("&reg;"), "®");
        HTML2TEXT.put(Pattern.compile("&#175;"), "¯");
        HTML2TEXT.put(Pattern.compile("&macr;"), "¯");
        HTML2TEXT.put(Pattern.compile("&#176;"), "°");
        HTML2TEXT.put(Pattern.compile("&deg;"), "°");
        HTML2TEXT.put(Pattern.compile("&#177;"), "±");
        HTML2TEXT.put(Pattern.compile("&plusmn;"), "±");
        HTML2TEXT.put(Pattern.compile("&#178;"), "²");
        HTML2TEXT.put(Pattern.compile("&sup2;"), "²");
        HTML2TEXT.put(Pattern.compile("&#179;"), "³");
        HTML2TEXT.put(Pattern.compile("&sup3;"), "³");
        HTML2TEXT.put(Pattern.compile("&#180;"), "´");
        HTML2TEXT.put(Pattern.compile("&acute;"), "´");
        HTML2TEXT.put(Pattern.compile("&#181;"), "µ");
        HTML2TEXT.put(Pattern.compile("&micro;"), "µ");
        HTML2TEXT.put(Pattern.compile("&#182;"), "¶");
        HTML2TEXT.put(Pattern.compile("&para;"), "¶");
        HTML2TEXT.put(Pattern.compile("&#183;"), "·");
        HTML2TEXT.put(Pattern.compile("&middot;"), "·");
        HTML2TEXT.put(Pattern.compile("&#184;"), "¸");
        HTML2TEXT.put(Pattern.compile("&cedil;"), "¸");
        HTML2TEXT.put(Pattern.compile("&#185;"), "¹");
        HTML2TEXT.put(Pattern.compile("&sup1;"), "¹");
        HTML2TEXT.put(Pattern.compile("&#186;"), "º");
        HTML2TEXT.put(Pattern.compile("&ordm;"), "º");
        HTML2TEXT.put(Pattern.compile("&#187;"), "»");
        HTML2TEXT.put(Pattern.compile("&raquo;"), "»");
        HTML2TEXT.put(Pattern.compile("&#188;"), "¼");
        HTML2TEXT.put(Pattern.compile("&frac14;"), "¼");
        HTML2TEXT.put(Pattern.compile("&#189;"), "½");
        HTML2TEXT.put(Pattern.compile("&frac12;"), "½");
        HTML2TEXT.put(Pattern.compile("&#190;"), "¾");
        HTML2TEXT.put(Pattern.compile("&frac34;"), "¾");
/*  -- Commented by RS to minimize quantity of actual Patterns
        HTML2TEXT.put(Pattern.compile("&#191;"), "¿");
        HTML2TEXT.put(Pattern.compile("&iquest;"), "¿");
        HTML2TEXT.put(Pattern.compile("&#192;"), "À");
        HTML2TEXT.put(Pattern.compile("&Agrave;"), "À");
        HTML2TEXT.put(Pattern.compile("&#193;"), "Á");
        HTML2TEXT.put(Pattern.compile("&Aacute;"), "Á");
        HTML2TEXT.put(Pattern.compile("&#194;"), "Â");
        HTML2TEXT.put(Pattern.compile("&#195;"), "Ã");
        HTML2TEXT.put(Pattern.compile("&Atilde;"), "Ã");
        HTML2TEXT.put(Pattern.compile("&#196;"), "Ä");
        HTML2TEXT.put(Pattern.compile("&Auml;"), "Ä");
        HTML2TEXT.put(Pattern.compile("&#197"), "Å");
        HTML2TEXT.put(Pattern.compile("&Aring;"), "Å");
        HTML2TEXT.put(Pattern.compile("&#198;"), "Æ");
        HTML2TEXT.put(Pattern.compile("&AElig;"), "Æ");
        HTML2TEXT.put(Pattern.compile("&#199;"), "Ç");
        HTML2TEXT.put(Pattern.compile("&Ccedil;"), "Ç");
        HTML2TEXT.put(Pattern.compile("&#200;"), "È");
        HTML2TEXT.put(Pattern.compile("&Egrave;"), "È");
        HTML2TEXT.put(Pattern.compile("&#201;"), "É");
        HTML2TEXT.put(Pattern.compile("&Eacute;"), "É");
        HTML2TEXT.put(Pattern.compile("&#202;"), "Ê");
        HTML2TEXT.put(Pattern.compile("&Ecirc;"), "Ê");
        HTML2TEXT.put(Pattern.compile("&#203;"), "Ë");
        HTML2TEXT.put(Pattern.compile("&#204;"), "Ì");
        HTML2TEXT.put(Pattern.compile("&Igrave;"), "Ì");
        HTML2TEXT.put(Pattern.compile("&#205;"), "Í");
        HTML2TEXT.put(Pattern.compile("&Iacute;"), "Í");
        HTML2TEXT.put(Pattern.compile("&#206;"), "Î");
        HTML2TEXT.put(Pattern.compile("&Icirc;"), "Î");
        HTML2TEXT.put(Pattern.compile("&#207;"), "Ï");
        HTML2TEXT.put(Pattern.compile("&Iuml;"), "Ï");
        HTML2TEXT.put(Pattern.compile("&#208;"), "Ð");
        HTML2TEXT.put(Pattern.compile("&ETH;"), "Ð");
        HTML2TEXT.put(Pattern.compile("&#209;"), "Ñ");
        HTML2TEXT.put(Pattern.compile("&Ntilde;"), "Ñ");
        HTML2TEXT.put(Pattern.compile("&#210;"), "Ò");
        HTML2TEXT.put(Pattern.compile("&Ograve;"), "Ò");
        HTML2TEXT.put(Pattern.compile("&#211;"), "Ó");
        HTML2TEXT.put(Pattern.compile("&Oacute;"), "Ó");
        HTML2TEXT.put(Pattern.compile("&#212;"), "Ô");
        HTML2TEXT.put(Pattern.compile("&Ocirc;"), "Ô");
        HTML2TEXT.put(Pattern.compile("&#213;"), "Õ");
        HTML2TEXT.put(Pattern.compile("&Otilde;"), "Õ");
        HTML2TEXT.put(Pattern.compile("&#214;"), "Ö");
        HTML2TEXT.put(Pattern.compile("&Ouml;"), "Ö");
        HTML2TEXT.put(Pattern.compile("&#215;"), "×");
        HTML2TEXT.put(Pattern.compile("&times;"), "×");
        HTML2TEXT.put(Pattern.compile("&#216;"), "Ø");
        HTML2TEXT.put(Pattern.compile("&Oslash;"), "Ø");
        HTML2TEXT.put(Pattern.compile("&#217;"), "Ù");
        HTML2TEXT.put(Pattern.compile("&Ugrave;"), "Ù");
        HTML2TEXT.put(Pattern.compile("&#218;"), "Ú");
        HTML2TEXT.put(Pattern.compile("&Uacute;"), "Ú");
        HTML2TEXT.put(Pattern.compile("&#219;"), "Û");
        HTML2TEXT.put(Pattern.compile("&Ucirc;"), "Û");
        HTML2TEXT.put(Pattern.compile("&#220;"), "Ü");
        HTML2TEXT.put(Pattern.compile("&Uuml;"), "Ü");
        HTML2TEXT.put(Pattern.compile("&#221;"), "Ý");
        HTML2TEXT.put(Pattern.compile("&Yacute;"), "Ý");
        HTML2TEXT.put(Pattern.compile("&#222;"), "Þ");
        HTML2TEXT.put(Pattern.compile("&THORN;"), "Þ");
        HTML2TEXT.put(Pattern.compile("&#223;"), "ß");
        HTML2TEXT.put(Pattern.compile("&szlig;"), "ß");
        HTML2TEXT.put(Pattern.compile("&#224;"), "à");
        HTML2TEXT.put(Pattern.compile("&agrave;"), "à");
        HTML2TEXT.put(Pattern.compile("&#225;"), "á");
        HTML2TEXT.put(Pattern.compile("&aacute;"), "á");
        HTML2TEXT.put(Pattern.compile("&#226;"), "â");
        HTML2TEXT.put(Pattern.compile("&acirc;"), "â");
        HTML2TEXT.put(Pattern.compile("&#227;"), "ã");
        HTML2TEXT.put(Pattern.compile("&atilde;"), "ã");
        HTML2TEXT.put(Pattern.compile("&#228;"), "ä");
        HTML2TEXT.put(Pattern.compile("&auml;"), "ä");
        HTML2TEXT.put(Pattern.compile("&#229;"), "å");
        HTML2TEXT.put(Pattern.compile("&aring;"), "å");
        HTML2TEXT.put(Pattern.compile("&#230;"), "æ");
        HTML2TEXT.put(Pattern.compile("&aelig;"), "æ");
        HTML2TEXT.put(Pattern.compile("&#231;"), "ç");
        HTML2TEXT.put(Pattern.compile("&ccedil;"), "ç");
        HTML2TEXT.put(Pattern.compile("&#232;"), "è");
        HTML2TEXT.put(Pattern.compile("&egrave;"), "è");
        HTML2TEXT.put(Pattern.compile("&#233;"), "é");
        HTML2TEXT.put(Pattern.compile("&eacute;"), "é");
        HTML2TEXT.put(Pattern.compile("&#234;"), "ê");
        HTML2TEXT.put(Pattern.compile("&ecirc;"), "ê");
        HTML2TEXT.put(Pattern.compile("&#235;"), "ë");
        HTML2TEXT.put(Pattern.compile("&euml;"), "ë");
        HTML2TEXT.put(Pattern.compile("&#236;"), "ì");
        HTML2TEXT.put(Pattern.compile("&igrave;"), "ì");
        HTML2TEXT.put(Pattern.compile("&#237"), "í");
        HTML2TEXT.put(Pattern.compile("&iacute;"), "í");
        HTML2TEXT.put(Pattern.compile("&#238;"), "î");
        HTML2TEXT.put(Pattern.compile("&icirc;"), "î");
        HTML2TEXT.put(Pattern.compile("&#239;"), "ï");
        HTML2TEXT.put(Pattern.compile("&iuml;"), "ï");
        HTML2TEXT.put(Pattern.compile("&#240;"), "ð");
        HTML2TEXT.put(Pattern.compile("&eth;"), "ð");
        HTML2TEXT.put(Pattern.compile("&#241;"), "ñ");
        HTML2TEXT.put(Pattern.compile("&ntilde;"), "ñ");
        HTML2TEXT.put(Pattern.compile("&#242;"), "ò");
        HTML2TEXT.put(Pattern.compile("&ograve;"), "ò");
        HTML2TEXT.put(Pattern.compile("&#243;"), "ó");
        HTML2TEXT.put(Pattern.compile("&oacute;"), "ó");
        HTML2TEXT.put(Pattern.compile("&#244;"), "ô");
        HTML2TEXT.put(Pattern.compile("&ocirc;"), "ô");
        HTML2TEXT.put(Pattern.compile("&#245;"), "õ");
        HTML2TEXT.put(Pattern.compile("&otilde;"), "õ");
        HTML2TEXT.put(Pattern.compile("&#246;"), "ö");
        HTML2TEXT.put(Pattern.compile("&ouml;"), "ö");
        HTML2TEXT.put(Pattern.compile("&#247;"), "÷");
        HTML2TEXT.put(Pattern.compile("&divide;"), "÷");
        HTML2TEXT.put(Pattern.compile("&#248;"), "ø");
        HTML2TEXT.put(Pattern.compile("&oslash;"), "ø");
        HTML2TEXT.put(Pattern.compile("&#249;"), "ù");
        HTML2TEXT.put(Pattern.compile("&ugrave;"), "ù");
        HTML2TEXT.put(Pattern.compile("&#250;"), "ú");
        HTML2TEXT.put(Pattern.compile("&uacute;"), "ú");
        HTML2TEXT.put(Pattern.compile("&#251;"), "û");
        HTML2TEXT.put(Pattern.compile("&ucirc;"), "û");
        HTML2TEXT.put(Pattern.compile("&#252;"), "ü");
        HTML2TEXT.put(Pattern.compile("&uuml;"), "ü");
        HTML2TEXT.put(Pattern.compile("&#253;"), "ý");
        HTML2TEXT.put(Pattern.compile("&yacute;"), "ý");
        HTML2TEXT.put(Pattern.compile("&#254;"), "þ");
        HTML2TEXT.put(Pattern.compile("&thorn;"), "þ");
        HTML2TEXT.put(Pattern.compile("&#255;"), "ÿ");
        HTML2TEXT.put(Pattern.compile("&#256;"), "Ā");
        HTML2TEXT.put(Pattern.compile("&#257;"), "ā");
        HTML2TEXT.put(Pattern.compile("&#258;"), "Ă");
        HTML2TEXT.put(Pattern.compile("&#259;"), "ă");
        HTML2TEXT.put(Pattern.compile("&#260;"), "Ą");
        HTML2TEXT.put(Pattern.compile("&#261;"), "ą");
        HTML2TEXT.put(Pattern.compile("&#262;"), "Ć");
        HTML2TEXT.put(Pattern.compile("&#263;"), "ć");
        HTML2TEXT.put(Pattern.compile("&#264;"), "Ĉ");
        HTML2TEXT.put(Pattern.compile("&#265;"), "ĉ");
        HTML2TEXT.put(Pattern.compile("&#266;"), "Ċ");
        HTML2TEXT.put(Pattern.compile("&#267;"), "ċ");
        HTML2TEXT.put(Pattern.compile("&#268;"), "Č");
        HTML2TEXT.put(Pattern.compile("&#269;"), "č");
        HTML2TEXT.put(Pattern.compile("&#270;"), "Ď");
        HTML2TEXT.put(Pattern.compile("&#271;"), "ď");
        HTML2TEXT.put(Pattern.compile("&#272;"), "Đ");
        HTML2TEXT.put(Pattern.compile("&#273;"), "đ");
        HTML2TEXT.put(Pattern.compile("&#274;"), "Ē");
        HTML2TEXT.put(Pattern.compile("&#275;"), "ē");
        HTML2TEXT.put(Pattern.compile("&#276;"), "Ĕ");
        HTML2TEXT.put(Pattern.compile("&#277;"), "ĕ");
        HTML2TEXT.put(Pattern.compile("&#278;"), "Ė");
        HTML2TEXT.put(Pattern.compile("&#279;"), "ė");
        HTML2TEXT.put(Pattern.compile("&#280;"), "Ę");
        HTML2TEXT.put(Pattern.compile("&#281;"), "ę");
        HTML2TEXT.put(Pattern.compile("&#282;"), "Ě");
        HTML2TEXT.put(Pattern.compile("&#283;"), "ě");
        HTML2TEXT.put(Pattern.compile("&#284;"), "Ĝ");
        HTML2TEXT.put(Pattern.compile("&#285;"), "ĝ");
        HTML2TEXT.put(Pattern.compile("&#286;"), "Ğ");
        HTML2TEXT.put(Pattern.compile("&#287;"), "ğ");
        HTML2TEXT.put(Pattern.compile("&#288;"), "Ġ");
        HTML2TEXT.put(Pattern.compile("&#289;"), "ġ");
        HTML2TEXT.put(Pattern.compile("&#290;"), "Ģ");
        HTML2TEXT.put(Pattern.compile("&#291;"), "ģ");
        HTML2TEXT.put(Pattern.compile("&#292;"), "Ĥ");
        HTML2TEXT.put(Pattern.compile("&#293;"), "ĥ");
        HTML2TEXT.put(Pattern.compile("&#294;"), "Ħ");
        HTML2TEXT.put(Pattern.compile("&#295;"), "ħ");
        HTML2TEXT.put(Pattern.compile("&#296;"), "Ĩ");
        HTML2TEXT.put(Pattern.compile("&#297;"), "ĩ");
        HTML2TEXT.put(Pattern.compile("&#298;"), "Ī");
        HTML2TEXT.put(Pattern.compile("&#299;"), "ī");
        HTML2TEXT.put(Pattern.compile("&#300;"), "Ĭ");
        HTML2TEXT.put(Pattern.compile("&#301;"), "ĭ");
        HTML2TEXT.put(Pattern.compile("&#302;"), "Į");
        HTML2TEXT.put(Pattern.compile("&#303;"), "į");
        HTML2TEXT.put(Pattern.compile("&#304;"), "İ");
        HTML2TEXT.put(Pattern.compile("&#305;"), "ı");
        HTML2TEXT.put(Pattern.compile("&#306;"), "Ĳ");
        HTML2TEXT.put(Pattern.compile("&#307;"), "ĳ");
        HTML2TEXT.put(Pattern.compile("&#308;"), "Ĵ");
        HTML2TEXT.put(Pattern.compile("&#309;"), "ĵ");
        HTML2TEXT.put(Pattern.compile("&#310;"), "Ķ");
        HTML2TEXT.put(Pattern.compile("&#311;"), "ķ");
        HTML2TEXT.put(Pattern.compile("&#312;"), "ĸ");
        HTML2TEXT.put(Pattern.compile("&#313;"), "Ĺ");
        HTML2TEXT.put(Pattern.compile("&#314;"), "ĺ");
        HTML2TEXT.put(Pattern.compile("&#315;"), "Ļ");
        HTML2TEXT.put(Pattern.compile("&#316;"), "ļ");
        HTML2TEXT.put(Pattern.compile("&#317;"), "Ľ");
        HTML2TEXT.put(Pattern.compile("&#318;"), "ľ");
        HTML2TEXT.put(Pattern.compile("&#319;"), "Ŀ");
        HTML2TEXT.put(Pattern.compile("&#320;"), "ŀ");
        HTML2TEXT.put(Pattern.compile("&#321;"), "Ł");
        HTML2TEXT.put(Pattern.compile("&#322;"), "ł");
        HTML2TEXT.put(Pattern.compile("&#323;"), "Ń");
        HTML2TEXT.put(Pattern.compile("&#324;"), "ń");
        HTML2TEXT.put(Pattern.compile("&#325;"), "Ņ");
        HTML2TEXT.put(Pattern.compile("&#326;"), "ņ");
        HTML2TEXT.put(Pattern.compile("&#327;"), "Ň");
        HTML2TEXT.put(Pattern.compile("&#328;"), "ň");
        HTML2TEXT.put(Pattern.compile("&#329;"), "ŉ");
        HTML2TEXT.put(Pattern.compile("&#330;"), "Ŋ");
        HTML2TEXT.put(Pattern.compile("&#331;"), "ŋ");
        HTML2TEXT.put(Pattern.compile("&#332;"), "Ō");
        HTML2TEXT.put(Pattern.compile("&#333;"), "ō");
        HTML2TEXT.put(Pattern.compile("&#334;"), "Ŏ");
        HTML2TEXT.put(Pattern.compile("&#335;"), "ŏ");
        HTML2TEXT.put(Pattern.compile("&#336;"), "Ő");
        HTML2TEXT.put(Pattern.compile("&#337;"), "ő");
        HTML2TEXT.put(Pattern.compile("&#338;"), "Œ");
        HTML2TEXT.put(Pattern.compile("&#339;"), "œ");
        HTML2TEXT.put(Pattern.compile("&#340;"), "Ŕ");
        HTML2TEXT.put(Pattern.compile("&#341;"), "ŕ");
        HTML2TEXT.put(Pattern.compile("&#342;"), "Ŗ");
        HTML2TEXT.put(Pattern.compile("&#343;"), "ŗ");
        HTML2TEXT.put(Pattern.compile("&#344;"), "Ř");
        HTML2TEXT.put(Pattern.compile("&#345;"), "ř");
        HTML2TEXT.put(Pattern.compile("&#346;"), "Ś");
        HTML2TEXT.put(Pattern.compile("&#347;"), "ś");
        HTML2TEXT.put(Pattern.compile("&#348;"), "Ŝ");
        HTML2TEXT.put(Pattern.compile("&#349;"), "ŝ");
        HTML2TEXT.put(Pattern.compile("&#350;"), "Ş");
        HTML2TEXT.put(Pattern.compile("&#351;"), "ş");
        HTML2TEXT.put(Pattern.compile("&#352;"), "Š");
        HTML2TEXT.put(Pattern.compile("&#353;"), "š");
        HTML2TEXT.put(Pattern.compile("&#354;"), "Ţ");
        HTML2TEXT.put(Pattern.compile("&#355;"), "ţ");
        HTML2TEXT.put(Pattern.compile("&#356;"), "Ť");
        HTML2TEXT.put(Pattern.compile("&#357;"), "ť");
        HTML2TEXT.put(Pattern.compile("&#358;"), "Ŧ");
        HTML2TEXT.put(Pattern.compile("&#359;"), "ŧ");
        HTML2TEXT.put(Pattern.compile("&#360;"), "Ũ");
        HTML2TEXT.put(Pattern.compile("&#361;"), "ũ");
        HTML2TEXT.put(Pattern.compile("&#362;"), "Ū");
        HTML2TEXT.put(Pattern.compile("&#363;"), "ū");
        HTML2TEXT.put(Pattern.compile("&#364;"), "Ŭ");
        HTML2TEXT.put(Pattern.compile("&#365;"), "ŭ");
        HTML2TEXT.put(Pattern.compile("&#366;"), "Ů");
        HTML2TEXT.put(Pattern.compile("&#367;"), "ů");
        HTML2TEXT.put(Pattern.compile("&#368;"), "Ű");
        HTML2TEXT.put(Pattern.compile("&#369;"), "ű");
        HTML2TEXT.put(Pattern.compile("&#370;"), "Ų");
        HTML2TEXT.put(Pattern.compile("&#371;"), "ų");
        HTML2TEXT.put(Pattern.compile("&#372;"), "Ŵ");
        HTML2TEXT.put(Pattern.compile("&#373;"), "ŵ");
        HTML2TEXT.put(Pattern.compile("&#374;"), "Ŷ");
        HTML2TEXT.put(Pattern.compile("&#375;"), "ŷ");
        HTML2TEXT.put(Pattern.compile("&#376;"), "Ÿ");
        HTML2TEXT.put(Pattern.compile("&#377;"), "Ź");
        HTML2TEXT.put(Pattern.compile("&#378;"), "ź");
        HTML2TEXT.put(Pattern.compile("&#379;"), "Ż");
        HTML2TEXT.put(Pattern.compile("&#380;"), "ż");
        HTML2TEXT.put(Pattern.compile("&#381;"), "Ž");
        HTML2TEXT.put(Pattern.compile("&#382;"), "ž");
        HTML2TEXT.put(Pattern.compile("&#383;"), "ſ");
        HTML2TEXT.put(Pattern.compile("&#340;"), "Ŕ");
        HTML2TEXT.put(Pattern.compile("&#341;"), "ŕ");
        HTML2TEXT.put(Pattern.compile("&#342;"), "Ŗ");
        HTML2TEXT.put(Pattern.compile("&#343;"), "ŗ");
        HTML2TEXT.put(Pattern.compile("&#344;"), "Ř");
        HTML2TEXT.put(Pattern.compile("&#345;"), "ř");
        HTML2TEXT.put(Pattern.compile("&#346;"), "Ś");
        HTML2TEXT.put(Pattern.compile("&#347;"), "ś");
        HTML2TEXT.put(Pattern.compile("&#348;"), "Ŝ");
        HTML2TEXT.put(Pattern.compile("&#349;"), "ŝ");
        HTML2TEXT.put(Pattern.compile("&#350;"), "Ş");
        HTML2TEXT.put(Pattern.compile("&#351;"), "ş");
        HTML2TEXT.put(Pattern.compile("&#352;"), "Š");
        HTML2TEXT.put(Pattern.compile("&#353;"), "š");
        HTML2TEXT.put(Pattern.compile("&#354;"), "Ţ");
        HTML2TEXT.put(Pattern.compile("&#355;"), "ţ");
        HTML2TEXT.put(Pattern.compile("&#356;"), "Ť");
        HTML2TEXT.put(Pattern.compile("&#577;"), "ť");
        HTML2TEXT.put(Pattern.compile("&#358;"), "Ŧ");
        HTML2TEXT.put(Pattern.compile("&#359;"), "ŧ");
        HTML2TEXT.put(Pattern.compile("&#360;"), "Ũ");
        HTML2TEXT.put(Pattern.compile("&#361;"), "ũ");
        HTML2TEXT.put(Pattern.compile("&#362;"), "Ū");
        HTML2TEXT.put(Pattern.compile("&#363;"), "ū");
        HTML2TEXT.put(Pattern.compile("&#364;"), "Ŭ");
        HTML2TEXT.put(Pattern.compile("&#365;"), "ŭ");
        HTML2TEXT.put(Pattern.compile("&#366;"), "Ů");
        HTML2TEXT.put(Pattern.compile("&#367;"), "ů");
        HTML2TEXT.put(Pattern.compile("&#368;"), "Ű");
        HTML2TEXT.put(Pattern.compile("&#369;"), "ű");
        HTML2TEXT.put(Pattern.compile("&#370;"), "Ų");
        HTML2TEXT.put(Pattern.compile("&#371;"), "ų");
        HTML2TEXT.put(Pattern.compile("&#372;"), "Ŵ");
        HTML2TEXT.put(Pattern.compile("&#373;"), "ŵ");
        HTML2TEXT.put(Pattern.compile("&#374;"), "Ŷ");
        HTML2TEXT.put(Pattern.compile("&#375;"), "ŷ");
        HTML2TEXT.put(Pattern.compile("&#376;"), "Ÿ");
        HTML2TEXT.put(Pattern.compile("&#377;"), "Ź");
        HTML2TEXT.put(Pattern.compile("&#378;"), "ź");
        HTML2TEXT.put(Pattern.compile("&#379;"), "Ż");
        HTML2TEXT.put(Pattern.compile("&#380;"), "ż");
        HTML2TEXT.put(Pattern.compile("&#381;"), "Ž");
        HTML2TEXT.put(Pattern.compile("&#382;"), "ž");
        HTML2TEXT.put(Pattern.compile("&#383;"), "ſ");
*/
        HTML2TEXT.put(Pattern.compile("&#38;"), "&");
        HTML2TEXT.put(Pattern.compile("&amp;"), "&");

        //remove double spaces
        HTML2TEXT.put(Pattern.compile("\\s{2,}+"), " ");
    }



    public static String chopBody(final String html) {
    	final int startBody = find_end(html_start, html);
    	final int startHtml = find_end(body_start, html);
    	int start = Math.max(startBody, startHtml);
    	if (start < 0) start = 0;
    	int endBody = find_start(html_end, html);
    	if (endBody < 0) endBody = html.length();
    	int endHtml = find_start(body_end, html);
    	if (endHtml < 0) endHtml = html.length();
    	final int end = Math.min(endHtml, endBody);
    	return html.substring(start, end);
	}

	private static int find_end(final Pattern pat, final String s) {
		final Matcher m = pat.matcher(s);
		if (!m.find()) return -1;
		return m.end();
	}

	private static int find_start(final Pattern pat, final String s) {
		final Matcher m = pat.matcher(s);
		if (!m.find()) return -1;
		return m.start();
	}

	
	private static final String [] htmlReplacements = new String [] {
		"&", "&amp;",
		"<", "&lt;",
		">", "&gt;",
	};
	
	/**
	 * Replace all non-html entities (&amp;, &lt; &gt;) with references (&amp;amp; &amp;lt; &amp;gt;) 
	 */
	public static String escape(String text) {
		for (int i=0; i<htmlReplacements.length/2; i++) {
			final int k = i << 1;
			text = text.replaceAll(htmlReplacements[k], htmlReplacements[k+1]);
		}
		return text;
	}

	/**
	 * Replace all non-html entity references (&amp;amp; &nbsp;, &amp;lt; &amp;gt;) with their textual values 
	 * (&amp;, &lt; &gt;) 
	 */
	public static String unescape(String html) {
		for (int i=0; i<htmlReplacements.length/2; i++) {
			final int k = i << 1;
			html = html .replaceAll(htmlReplacements[k+1], htmlReplacements[k]);
		}
		return html.replaceAll("&nbsp;", " ");
	}

	/**
	 * Remove all tags from html so that it will contain only text. All tags will be replaced with spaces.
     * try also {@link HtmlUtil#html2text}
	 */
	public static String stripTags(final String html) {
		return html.replaceAll("<.*?>", " ");
	}

	/**
	 * Replaces multiple whitespaces with one, removes spaces where they are not required (for example, after the tag). 
	 * therefore can be used to produce "canonical" html that can be easilly compared to other "canonical" html 
	 */
	public static String canonize(final String html) {
		return html.replaceAll("\\s+", " ").replaceAll(">\\s+", ">").replaceAll("\\s+<", "<")
			.replaceAll("\\s*&nbsp;\\s*", "&nbsp;").trim();
	}


    public static String html2text(CharSequence htmlInput) {
        String text = htmlInput.toString();
        for (Map.Entry<Pattern, String> me : HTML2TEXT.entrySet()) {
            //create a Matcher
            Matcher m = me.getKey().matcher(text);
            //replace all
            text = m.replaceAll(me.getValue());
        }
        return text.trim();
    }

    private static final Set<String> CLEAN_TAGS;
    
    static {
    	CLEAN_TAGS = new HashSet<String>(5);
    	CLEAN_TAGS.add("b");
    	CLEAN_TAGS.add("br");
    	CLEAN_TAGS.add("i");
    	CLEAN_TAGS.add("u");
    	CLEAN_TAGS.add("strong");
    }
    
    private static final Pattern TAG = Pattern.compile("<(?:[/]??)(\\w+).*?(?:>|$)");
    
	public static String clean(final String html) {
		final StringBuilder b = new StringBuilder(html.length());
		final Matcher m = TAG.matcher(html);
		
		int pos = 0;

		while (m.find()) {
			b.append(html.substring(pos, m.start()));

			final String tagName = m.group(1);
			final boolean allowedTag = CLEAN_TAGS.contains(tagName.toLowerCase());
			
			String escaped = m.group();
			if (!allowedTag) escaped = escape(escaped);

			b.append(escaped);
			pos = m.end();
		}
		
		b.append(html.substring(pos));
		
		return b.toString();
	}


}
