package org.builtu;

import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.builtu.config.module.Config;
import org.builtu.config.module.ConfigUtil;
import org.builtu.config.module.Configurable;


public class Aggregator implements StatInput, Configurable {
	
	public static final String MODULE="webstat-core";
	
	private int interval=1000;
	private long sessionTimeout=1000L * 60L * 30L; // 30 min by default

	/**
	 * @return aggregation interval in milliseconds
	 */
	public int getInterval() {
		return interval;
	}

	/**
	 * set aggregation interval in milliseconds
	 */
	public void setInterval(int interval) {
		this.interval = interval;
	}
	
	/**
	 * @return session timeout in milliseconds
	 */
	public long getSessionTimeout() {
		return sessionTimeout;
	}

	/**
	 * set session timeout in milliseconds
	 */
	public void setSessionTimeout(long sessionTimeout) {
		this.sessionTimeout = sessionTimeout;
	}
	
	public Aggregator() {
		ConfigUtil.configure(MODULE, this);
	}

	@Override public void configure(final String moduleName, final Config config) {
		final int hintExpectedSessionsPerThreadPerSecond = 
			config.getInt(MODULE+".hints.expectedSessionsPerThreadPerSecond");
		final int hintExpectedThreads =
			config.getInt(MODULE+".hints.hintExpectedThreads");
		
		init(hintExpectedSessionsPerThreadPerSecond, hintExpectedThreads);
		
		final int interval = config.getInt(MODULE+".interval");
		this.setInterval(interval);
		
		final long sessionTimeout = config.getInt(MODULE+".sessionTimeout");
		this.setSessionTimeout(sessionTimeout);
	}

	private ThreadLocal<Map<String, Long>> threadLocalSessions;
	private ConcurrentHashMap<String, Long> globalSessions;
	private int expectedThreadLocalSize;
	
	private void init (final int hintExpectedSessionsPerThreadPerSecond, final int hintExpectedThreads) {
		this.expectedThreadLocalSize = hintExpectedSessionsPerThreadPerSecond* interval/1000;
		threadLocalSessions = new ThreadLocal<Map<String, Long>>();
		final int expectedGlobalSessions = expectedThreadLocalSize * hintExpectedThreads;
		globalSessions = new ConcurrentHashMap<String, Long>(expectedGlobalSessions);
	}
	
	private Map<String, Long> getLocalSessions() {
		Map<String, Long> res = threadLocalSessions.get();
		if (res == null) {
			synchronized(threadLocalSessions) {
				res = threadLocalSessions.get();
				if (res != null) return res;
				res = new HashMap<String, Long>(expectedThreadLocalSize);
				threadLocalSessions.set(res);
			}
		}
		return res;
	}

	public final Set<StatListener> listeners = new HashSet<StatListener>(1);
	
	private final ThreadLocal<Long> prevNotification = new ThreadLocal<Long>();

	@Override public void keepAlive(final String sessionId) {
		final long now = System.currentTimeMillis();
		getLocalSessions().put(sessionId, now);
		notifyIfNeeded(now);
	}
	
	public void commit() {
		prevNotification.set(0L);
		final long now = System.currentTimeMillis();
		notifyIfNeeded(now);
	}

	private void notifyIfNeeded(final long now) {
		if (prevNotification.get() == null) {
			prevNotification.set(now);
			return;
		}
		if (now - prevNotification.get() >= interval) {
			aggregate(now);
			prevNotification.set(now);
		}
	}
	
	private long notificationStarted;
	private final Object notificationLock = new Object();
	
	private void aggregate(final long now) {
		updateGlobalSessions(now);
		boolean notificationStalled = false;
		if (notificationStarted > 0) {
			// notification is considered stalled (notification thread died or other shit happened)
			// if notification was not finished 10 intervals ago
			notificationStalled = ((now - notificationStarted) / interval) > 10;
			if (!notificationStalled) return;
		}
		synchronized (notificationLock) {
			if (notificationStarted>0) {
				notificationStalled = ((now - notificationStarted) / interval) > 10;
				if (!notificationStalled) return;
			}
			notificationStarted = now;
			try {
				notifyListeners(now);
			} finally {
				notificationStarted = 0;
			}
		}
	}

	private void notifyListeners(final long now) {
		removeDeadGlobalSessions(now);
		
		final int liveSessionsCount = globalSessions.size();
		final Date from = new Date(prevNotification.get());
		final Date to = new Date(now);

		for (final StatListener l : listeners) {
			l.listen(from, to, liveSessionsCount);
		}
	}

	private void updateGlobalSessions(long now) {
		final Map<String, Long> sessions = getLocalSessions();
		final Iterator<Map.Entry<String, Long>> it = sessions.entrySet().iterator();
		while (it.hasNext()) {
			final Map.Entry<String, Long> e = it.next();
			final String id = e.getKey();
			final long date = e.getValue();
			if (now - date >= sessionTimeout) {
				it.remove();
			} else {
				updateGlobalDate(id, date);
			}
		}
	}

	private void updateGlobalDate(final String id, final long date) {
		Long old = globalSessions.putIfAbsent(id, date);
		
		if (old == null) return;
		
		boolean noUpdateIsNessesaryWhenLocalDateIsLessThanGlobal = date < old; 
		if (noUpdateIsNessesaryWhenLocalDateIsLessThanGlobal) return;
		
		while (!globalSessions.replace(id, old, date)) {
			old = globalSessions.get(id);
			noUpdateIsNessesaryWhenLocalDateIsLessThanGlobal = old != null && date < old;
			if (noUpdateIsNessesaryWhenLocalDateIsLessThanGlobal) return;
		}
	}

	private void removeDeadGlobalSessions(final long now) {
		final Iterator<Map.Entry<String, Long>> it = globalSessions.entrySet().iterator();
		while (it.hasNext()) {
			final Map.Entry<String, Long> e = it.next();
			final String id = e.getKey();
			final Long date = e.getValue();
			if (date == null) continue;
			if (now - date >= sessionTimeout) {
				globalSessions.remove(id, date);
			}
		}
	}

	private static volatile Aggregator instance;
	
	public static Aggregator getInstance() {
		if (instance != null) return instance;
		synchronized(Aggregator.class) {
			if (instance != null) return instance;
			instance = new Aggregator();
		}
		return instance;
	}
}
